#!/bin/bash
# lessfsconfig.sh v0.0.2
# This script will create a lessfs configuration file


# Set Default Values

cfg_version=0.0.2
cfg_METAQSIZE=2048
fs_block_size=64
fs_debug=3
fs_hashlen=24
fs_blockdata_path=/data/dta
fs_blockdata_file=blockdata.dta
fs_file_io=true
fs_blockdata_bs_gb=600
fs_fileblock_bs_gb=6411
fs_metadata_path=/data/mta
fs_max_files=524288
fs_cache_size_mb=128
fs_max_threads=2
fs_dynamic_defrag=true
fs_sync_relax=1
fs_encrypt_data=false
fs_encrypt_meta=false
fs_password=1234
fs_address=127.0.0.1
fs_port=100
fs_min_space_free=10
fs_min_space_clean=15
fs_clean_exec=false


# functions
get_param(){
  local pvalue=`awk '{ sub("\r", ""); print }' "$1"|awk '$1~s{f++}f&&$1~p{split($1,a,m);gsub("[\t ]*$","",a[2]);print a[2];exit}' s=".*" p="^[\t ]*$2" m="^[\t ]*$2[\t ]*=[\t ]*" FS="\$^"`
  if [ -n $3 ] && [ -z $pvalue ]; then
      eval $2="$3"
    else
      eval $2="$pvalue"
  fi
}

check_path(){
  if [[ "$2" =~ ^/ ]]
    then
      if [[ "$2" =~ "(/$)|(/\.)|(/-)" ]] && [ "$2" != "/" ]
        then
          parse_error[$((${#parse_error[*]}+1))]="\"$2\" is an invalid path.  Must not end with \"/\" or contain \"/.\" or \"/-\"."
        else
          eval $1="$2"
      fi
    else
      parse_error[$((${#parse_error[*]}+1))]="\"$2\" is an invalid path.  Must start with \"/\"."
  fi
}

check_file(){
  if [[ "$2" =~ (^\.)|(/)|(-) ]]
    then
      parse_error[$((${#parse_error[*]}+1))]="\"$2\" is an invalid filename.  Must not contain a \"/\", or start with \".\" or \"-\"."
    else
      eval $1="$2"
  fi
}

check_number(){
  if [[ "$2" =~ ^[0-9]+$ ]] && [ $2 -ge $3 ] && [ $2 -le $4 ]
    then
      eval $1="$2"
    else
      parse_error[$((${#parse_error[*]}+1))]="\"$2\" is invalid for $c. Use an integer number from $3 through $4."
  fi
}

check_input(){
  local variable_name=$1
  shift
  local parm_name=$1
  shift
  local new_value=$1
  shift
  local match=false
  while [ "$1" ]; do
    if [ "$new_value" = "$1" ]
      then
        match=true
    fi
    local option_list+=" $1"
    shift
  done
  if [ "$match" = "true" ]
    then
      eval $variable_name="$new_value"
    else
      parse_error[$((${#parse_error[*]}+1))]="\"$new_value\" is invalid for $parm_name. Valid options are:$option_list."
  fi
}

apnd_cfg(){
  local parm=$1
  shift
  if [ ! -z $1 ]; then
    parm+="="
    parm+=$1
    shift
    if [ ! -z $1 ]; then
      local commentout=$1
    fi
  fi
  string=""
  if [ "$commentout" = "true" ]; then
    string+="#"
  fi
  string+="$parm"
  echo "$string" >> $outfile
}

# Parameter parsing
cmd_count=0
while [ "${1:0:1}" == '-' ]; do
  case "$1" in

    --cache_size) c=$1; shift; check_number fs_cache_size_mb $1 1 16384 $c;;

    --debug) c=$1; shift; check_number fs_debug $1 0 5 $c;;

    --block_size) c=$1; shift; check_input fs_block_size $c $1 4 8 16 32 64 128;;

    "-h" | "--help" | "-?") show_help=true;;

    --force) overwrite_file=true;;

    --blockdata_path) shift; check_path fs_blockdata_path $1;;

    --metadata_path) shift; check_path fs_metadata_path $1;;

    --blockdata_file) shift; check_file fs_blockdata_file $1;;

    --blockdata_size) c=$1; shift; check_number fs_blockdata_bs_gb $1 1 1048576 $c;;

    --fileblock_size) c=$1; shift; check_number fs_fileblock_bs_gb $1 1 1048576 $c;;

    --file_io) c=$1; shift; check_input fs_file_io $c $1 "true" "false";;

    --dynamic_defrag) c=$1; shift; check_input fs_dynamic_defrag $c $1 "true" "false";;

    --max_files) c=$1; shift; check_number fs_max_files $1 1024 1073741824 $c;;

    --max_threads) c=$1; shift; check_number fs_max_threads $1 1 64 $c;;

    --sync_relay) c=$1; shift; check_number fs_sync_relax $1 0 2 $c;;

    --check) read_file=true;;

    *) parse_error[$((${#parse_error[*]}+1))]="\"$i\" is an unknown option. Use --help for more information.";;

  esac
  shift
  cmd_count=$(($cmd_count+1))
done
outfile=$1
shift
if [ $1 ]
  then
    parse_error[$((${#parse_error[*]}+1))]="\"$1\" is invalid. All options must be specified before the outfile."
fi

# check for output file
if [ -z $outfile ]
  then
    if [ $cmd_count -gt 0 ] && [ -z $show_help ]
      then
        parse_error[$((${#parse_error[*]}+1))]="You must enter an ouput file. Use --help for more information."
      else
        show_help=true
    fi
fi

# check status of output file
if [ -e "$outfile" ] && [ -z $show_help ]
  then
    if [ -d "$outfile" ]
      then
        parse_error[$((${#parse_error[*]}+1))]="\"$outfile\" is a directory."
      elif [ -z $overwrite_file ] && [ -z $read_file ]; then
        parse_error[$((${#parse_error[*]}+1))]="\"$outfile\" already exists.  Use --force to overwrite."
      elif [ ! -w "$outfile" ] && [ -z $read_file ]; then
        parse_error[$((${#parse_error[*]}+1))]="\"$outfile\" does not have write permissions."
    fi
  elif [ "$read_file" = "true" ] && [ -z $show_help ]; then
    parse_error[$((${#parse_error[*]}+1))]="\"$outfile\" does not exist.  Cannot open for reading"
fi

# output any error
if [ ${#parse_error[@]} -gt 0 ]
  then
  for i in ${!parse_error[*]}
    do
    echo "Error: ${parse_error[$i]}" >&2
  done
  exit 3
fi

# display help
if [ $show_help ]
  then
  echo "lessfsconfig.sh v$cfg_version"
  echo ""
  echo "Creates a new lessfs configuration file"
  echo "---------------------------------------------"
  echo "Usage: $0 [options] output_file"
  echo ""
  echo "  --help"
  echo "    Show this help screen."
  echo ""
  echo "  --check"
  echo "    Display statistics on an existing configuration file."
  echo "    No changes will be made to configuration file."
  echo ""
  echo "  --force"
  echo "    Force script to overwite any existing file.  Use with caution."
  echo ""
  echo "  --debug"
  echo "    Valid options are 0..5."
  echo "    Default: 3"
  echo "    0  Disable logging with the exception of critical errors."
  echo "    1  Enable logging of critical errors and warnings."
  echo "    2  Enable logging of critical errors, warnings and informational messages."
  echo "    3  Enable logging of critical errors, warnings, informational messages and debug messages."
  echo "    >3 Enable logging of critical errors, warnings, informational messages and debug and other messages."
  echo ""
  echo "  --block_size"
  echo "    Given in KB, sets the size of blocks to use."
  echo "    Default: 64"
  echo "    Valid options: 4, 8, 16, 32, 64, 128"
  echo "    Note:"
  echo "    A larger block size can dramatically increase performance.  64 or 128 is suggested."
  echo ""
  echo "  --blockdata_path"
  echo "    Set path for storage of block data."
  echo "    Default: /data/dta"
  echo ""
  echo "  --blockdata_file"
  echo "    Set name of file used for file_io storage (not used if TokyoCabinet storage is selected.)"
  echo "    Default: blockdata.dta"
  echo ""
  echo "  --blockdata_size"
  echo "    Given in GB, configures lessfs for the maximum amount of storage you expect to use, after deduplication."
  echo "    Default: 600"
  echo "    Note:"
  echo "    This setting is used to calculate the following options in the lssfs configuration file:"
  echo "    BLOCKDATA_BS (not used if file_io is selected)"
  echo "    FREELIST_BS (not used if TokyoCabinet is selected)"
  echo "    BLOCKUSAGE_BS"
  echo "    The calculation used is:"
  echo "    ((blockdata_size*1024*1024*1024)/(block_size*1024))*2"
  echo "    The settings is approximate.  It does not hurt to shoot high, if you have enough RAM."
  echo "    Setting this too low can cause serious issues."
  echo ""
  echo "  --fileblock_size"
  echo "    Given in GB, configures lessfs for the maximum amount of storage you expect to use, before deduplication."
  echo "    Default: 6411 (6 TB)"
  echo "    Note:"
  echo "    This setting is used to calculate the following option in the lssfs configuration file:"
  echo "    FILEBLOCK_BS"
  echo "    The calculation used is:"
  echo "    ((fileblock_size*1024*1024*1024)/(block_size*1024))*2"
  echo "    The settings is approximate.  It does not hurt to shoot high, if you have enough RAM."
  echo "    Setting this too low can cause serious issues."
  echo ""
  echo "  --file_io"
  echo "    Select file_io or TokyoCabinet for storage."
  echo "    Default: true"
  echo "    Valid options:"
  echo "    true    Use file_io for block storage"
  echo "    false   Use TokyoCabinet for block storage"
  echo "    Note:"
  echo "    file_io performs faster than TokyoCabinet, but will not shrink in size when files are deleted."
  echo ""
  echo "  --metadata_path"
  echo "    Sets path to the database files for file meta data."
  echo "    Default: /data/mta"
  echo "    Note:"
  echo "    This single setting sets the lessfs configuration paths for:"
  echo "    BLOCKUSAGE_PATH, DIRENT_PATH, FILEBLOCK_PATH, "
  echo "    META_PATH, HARDLINK_PATH, SYLINK_PATH and FREELIST_PATH"
  echo "    For best performance, these should be located on a fast storage device."
  echo ""
  echo "  --max_files"
  echo "    Sets the maximum number of files lessfs can store**."
  echo "    Default: 524288"
  echo "    Note:**"
  echo "    This single setting sets the lessfs configuration for:"
  echo "    DIRENT_BS, META_BS, HARDLINK_BS and SYMLINK_BS"
  echo "    Using the calculation: (maxfiles*2)"
  echo "    This means that a max_files setting of 524288 allows you to have:"
  echo "    524,288 Directory entries, 524,288 Meta entries, 524,288 Hardlinks and 524,288 Symlinks"
  echo ""
  echo "  --cache_size"
  echo "    The cachesize megabytes that lessfs is allowed to use as write cache."
  echo "    Default: 128"
  echo ""
  echo "  --max_threads"
  echo "    Default: 2"
  echo "    Do  not  set MAX_THREADS higher then the number of CPU cores available in"
  echo "    the system. Lessfs will use MAX_THREADS CPU cores for LZO or QUICKLZ"
  echo "    compression to compress MAX_THREADS data blocks in parallel. The performance"
  echo "    will suffer when this number is set higher then the number of CPU cores that"
  echo "    are available in the system."
  echo ""
  echo "  --dynamic_defrag"
  echo "    Enable tokyocabinets automatic defragmentation feature."
  echo "    Default: true"
  echo ""
  echo "  --sync_relax"
  echo "    Default: 1"
  echo "    Valid options are : 0, 1 or 2"
  echo "    0  Flush  all caches in lessfs for an inode and sync the tokyocabinet"
  echo "       databases to disk when fsync is called for an inode."
  echo "    1  Do not sync the tokyocabinet databases to the disk when fsync is called"
  echo "       on an inode.  The inode data will be written directly to the databases."
  echo "       In case of a crash the databases themselves might not be committed to"
  echo "       disk.  This feature improves some types of I/O and is especially useful"
  echo "       with NFS.  There is a trade-off between more speed and the chance of"
  echo "       possible loss of data. Use with caution."
  echo "    2  Living on the edge.  Do not flush the caches in lessfs and do not sync"
  echo "       the tokyocabinet databases to disk.  Use with extreme caution."
  echo ""
  echo ""
  exit 3
fi

if [ -z $read_file ]; then
# Create config file
echo "# Automatically generated by lessfsconfig.sh" > $outfile
# check to make sure it is there
if [ ! -f "$outfile" ]
  then
    echo "Error:  Unable to create file \"$outfile\"." >&2
    exit 3
fi

# setup, and append the rest of the config file
BLOCKDATA_BS=$(((($fs_blockdata_bs_gb*1024*1024*1024)/($fs_block_size*1024))*2))
FILEBLOCK_BS=$(((($fs_fileblock_bs_gb*1024*1024*1024)/($fs_block_size*1024))*2))
META_BS=$(($fs_max_files*2))

apnd_cfg "# Enable informational messages about compression."
apnd_cfg DEBUG $fs_defug
apnd_cfg "# The (chopped) hashlen in bytes, minimum is 20."
apnd_cfg "HASHLEN = 24"
apnd_cfg BLOCKDATA_IO_TYPE file_io `[ ! "$fs_file_io" = "true" ] && echo "true"`
apnd_cfg BLOCKDATA_PATH "$fs_blockdata_path/$fs_blockdata_file" `[ ! "$fs_file_io" = "true" ] && echo "true"`
apnd_cfg BLOCKDATA_PATH $fs_blockdata_path $fs_file_io
apnd_cfg BLOCKDATA_BS $BLOCKDATA_BS $fs_file_io
apnd_cfg "#"
apnd_cfg BLOCKUSAGE_PATH $fs_metadata_path
apnd_cfg BLOCKUSAGE_BS $BLOCKDATA_BS
apnd_cfg "#"
apnd_cfg DIRENT_PATH $fs_metadata_path
apnd_cfg DIRENT_BS $META_BS
apnd_cfg "#"
apnd_cfg FILEBLOCK_PATH $fs_metadata_path
apnd_cfg FILEBLOCK_BS $FILEBLOCK_BS
apnd_cfg "#"
apnd_cfg META_PATH $fs_metadata_path
apnd_cfg META_BS $META_BS
apnd_cfg "#"
apnd_cfg HARDLINK_PATH $fs_metadata_path
apnd_cfg HARDLINK_BS $META_BS
apnd_cfg "#"
apnd_cfg SYMLINK_PATH $fs_metadata_path
apnd_cfg SYMLINK_BS $META_BS
apnd_cfg "#"
apnd_cfg "# The freelist database is only used"
apnd_cfg "# with the  file_io backend"
apnd_cfg "#"
apnd_cfg FREELIST_PATH $fs_metadata_path
apnd_cfg FREELIST_BS $BLOCKDATA_BS
apnd_cfg "#"
apnd_cfg "# CACHESIZE in MB"
apnd_cfg CACHESIZE $fs_cache_size_mb
apnd_cfg "# Flush data to disk after X seconds."
apnd_cfg "COMMIT_INTERVAL=30"
apnd_cfg "#"
apnd_cfg "LISTEN_IP=127.0.0.1"
apnd_cfg "LISTEN_PORT=100"
apnd_cfg "# Not more then 2 on most machines."
apnd_cfg MAX_THREADS $fs_max_threads
apnd_cfg "#DYNAMIC_DEFRAGMENTATION on or off, default is off."
apnd_cfg DYNAMIC_DEFRAGMENTATION on `[ ! "$fs_dynamic_defrag" = "true" ] && echo "true"`
apnd_cfg "COREDUMPSIZE=256000000"
apnd_cfg "# Consider SYNC_RELAX=1 or SYNC_RELAX=2 when exporting lessfs with NFS."
apnd_cfg SYNC_RELAX $fs_sync_relax
apnd_cfg "# Requires openssl and lessfs has to be configured with --with-crypto"
apnd_cfg "ENCRYPT_DATA=off"
apnd_cfg "# ENCRYPT_META on or off, default is off"
apnd_cfg "# Requires ENCRYPT_DATA=on and is otherwise ignored."
apnd_cfg "ENCRYPT_META=on"
apnd_cfg "#PASSWORD=1234"
apnd_cfg "MIN_SPACE_FREE=10"
apnd_cfg "#MIN_SPACE_CLEAN=15"
apnd_cfg "#CLEAN_PROGRAM=/tmp/clean.sh"
apnd_cfg "# Highly recommended."
apnd_cfg "ENABLE_TRANSACTIONS=on"
apnd_cfg "# Select a blocksize to fit your needs."
apnd_cfg "# Most people should select 65536"
apnd_cfg BLKSIZE $(($fs_block_size*1024))
apnd_cfg "#BLKSIZE=131072"
apnd_cfg "#BLKSIZE=65536"
apnd_cfg "#BLKSIZE=32768"
apnd_cfg "#BLKSIZE=16384"

echo ""
echo "Created file: $outfile"

fi

#Read back valus from configuration file to verify them

get_param $outfile BLKSIZE 0
get_param $outfile BLOCKDATA_IO_TYPE
get_param $outfile BLOCKDATA_BS 0
[ "$BLOCKDATA_IO_TYPE" = "file_io" ] && BLOCKDATA_BS=0
get_param $outfile BLOCKUSAGE_BS 0
get_param $outfile DIRENT_BS 0
get_param $outfile FILEBLOCK_BS 0
get_param $outfile META_BS 0
get_param $outfile HARDLINK_BS 0
get_param $outfile SYMLINK_BS 0
get_param $outfile FREELIST_BS 0
[ ! "$BLOCKDATA_IO_TYPE" = "file_io" ] && FREELIST_BS=0
get_param $outfile BLKSIZE 0
get_param $outfile CACHESIZE 0

# add up how much RAM this configuration will use

req_blockdata=$(((($BLOCKDATA_BS+$BLOCKUSAGE_BS+$FREELIST_BS)/1024)/1024))
req_fileblock=$(((($FILEBLOCK_BS)/1024)/1024))
req_metadata=$(((($DIRENT_BS+$META_BS+$HARDLINK_BS+$SYMLINK_BS)/1024)/1024))
if [ $BLOCKUSAGE_BS -lt $BLOCKDATA_BS ] || [ "$BLOCKDATA_IO_TYPE" = "file_io" ]; then
    min_block=$BLOCKUSAGE_BS
  else
    min_block=$BLOCKDATA_BS
fi
fs_blockdata_bs_gb=$(((((($min_block/2)*$BLKSIZE)/1024)/1024)/1024))
fs_fileblock_bs_gb=$(((((($FILEBLOCK_BS/2)*$BLKSIZE)/1024)/1024)/1024))
req_blockbuffer=$(((($BLKSIZE/1024)*(cfg_METAQSIZE*2)/1024)))
fs_block_size=$(($BLKSIZE/1024))
req_total=$(($req_blockdata+$req_fileblock+$req_metadata+$req_blockbuffer+$CACHESIZE))

echo ""
echo "Statistics for: $outfile"
echo ""
echo "This configuration will need $req_total MB of available RAM to operate well."
echo "  $CACHESIZE MB for cache"
echo "  $req_blockdata MB for blockdata database buckets"
echo "  $req_fileblock MB for fileblock database buckets"
echo "  $req_metadata MB for metadata database buckets"
echo "  $req_blockbuffer MB for block write buffer"
echo ""
echo "This configuration will:"
echo "  Use up to $fs_blockdata_bs_gb GB of physical space before suffering performance and stability issues."
echo "  Supply up to $fs_fileblock_bs_gb GB of logical, pre-deduplicated, storage before suffering performance and stability issues."
echo ""
[ $fs_block_size -lt 64 ] && echo "You are using a block size of $fs_block_size KB.  Block sizes less than 64 KB may result in slower performance."
echo ""
